# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.stable_xl_refiner.text2image import (
    StableXLRefinerForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_xl_refiner.image2image import (
    StableXLRefinerForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_xl_refiner.inpainting import (
    StableXLRefinerForImageInpaintingPipeline,
)
