# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.controlnet_xl.text2image import (
    ControlNetXLForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet_xl.image2image import (
    ControlNetXLForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet_xl.inpainting import (
    ControlNetXLForImageInpaintingPipeline,
)
