# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.beit.modeling import (
    BeitForImageClassification,
)
from unitorch.models.beit.processing import BeitProcessor
