# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.stable.text2image import (
    StableForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable.image2image import (
    StableForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable.inpainting import StableForImageInpaintingPipeline
from unitorch.cli.pipelines.stable.resolution import StableForImageResolutionPipeline
