# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.peft.bloom import BloomLoraForGenerationPipeline
from unitorch.cli.pipelines.peft.mistral import MistralLoraForGenerationPipeline
